/**************************************************************************
 * Copyright (C) 2010 Yinhao Jiang <Jiang.Yinhao@gmail.com>
 * All rights reserved.
 * Created by Southeast University RoboCup.
 **************************************************************************
 */

package seu.entities;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import seu.states.RoadState;
import seu.utilities.AgentUtil;
import seu.utilities.ConstantConditions;

/**
 * @author Yinhao Jiang
 * @since 2010.5.18
 */
// //
public class Path {
	@SuppressWarnings("static-access")
	public Path(int id, StandardWorldModel world) {
		this.id = id;
		headNode = null;
		tailNode = null;
		roads = new LinkedList<Road>();
		nodes = new HashSet<Node>();
		connectedBuildings = new HashSet<Building>();
		this.world = world;
	}

	@SuppressWarnings("static-access")
	public Path(int id, HashSet<Road> srcRoads, StandardWorldModel world) {
		Path path = extractRoadsToOnePath_remove(id, srcRoads, world);
		this.id = path.id;
		this.headNode = path.headNode;
		this.tailNode = path.tailNode;
		this.roads = path.roads;
		this.nodes = path.nodes;
		this.connectedBuildings = path.connectedBuildings;
		this.world = world;
	}

	public int id;

	private static StandardWorldModel world;

	private LinkedList<Road> roads;

	private HashSet<Node> nodes;

	private HashSet<Building> connectedBuildings;

	private Road headNode = null;

	private Road tailNode = null;

	private ArrayList<EntityID> headToTailRoute = null;

	private ArrayList<EntityID> tailToHeadRoute = null;

	private Road m_center;

	public Road getCenter() {
		if (m_center == null) {
			m_center = roads.get(roads.size() / 2);
		}
		return m_center;
	}

	public String toString() {
		return "Path [id = " + id + "  center = " + getCenter() + "]";
	}

	public int headIndex = -1;

	public int tailIndex = -1;

	private boolean oneEndDead = false;

	public void setOneEndDead() {
		oneEndDead = true;
	}

	public boolean isOneEndDead() {
		return oneEndDead;
	}

	public boolean isBlocked() {
		return ConstantConditions.PASSABLE_C.extract(roads).size() < roads
				.size();
	}

	private boolean m_allRoadsKnown = false;

	private ArrayList<Edge> edges;

	public boolean allRoadsKnown() {
		if (m_allRoadsKnown) {
			return m_allRoadsKnown;
		}
		for (Road rd : roads) {
			if (rd.getState().getCondition() == RoadState.CONDITION_UNKNOWN) {
				return false;
			}
		}
		m_allRoadsKnown = true;
		return m_allRoadsKnown;
	}

	public boolean isPassed() {
		return ConstantConditions.HAS_NOT_BEEN_PASSED_C.extract(roads).size() < roads
				.size();
	}

	public boolean isOnPath(Node n1, Node n2) {
		return (nodes.contains(n1) && nodes.contains(n2));
	}

	public boolean isOnPath(StandardEntity mlo) {
		if (mlo == null)
			return false;
		if (edges == null) {
			edges = new ArrayList<Edge>();
			for (Road rd : roads)
				for (Edge eg : rd.getEdges())
					if (eg.isPassable())
						edges.add(eg);
		}
		return (edges.contains(mlo) || roads.contains(mlo));
	}

	public boolean isNeighbourPath(Path path) {
		return (nodes.contains(path.getHeadNode()) || nodes.contains(path
				.getTailNode()));
	}

	static public Path extractRoadsToOnePath_remove(int id,
			HashSet<Road> srcRoads, StandardWorldModel world) {
		if (srcRoads.isEmpty()) {
			return null;
		}
		HashSet<Road> resultSet = new HashSet<Road>();
		Path resultPath = null;
		HashSet<Road> headNeighborRoads = null;
		HashSet<Road> tailNeighborRoads = null;
		List<Road> temp = new ArrayList<Road>(srcRoads);
		Road oneOfSrcRoads = temp.get(0);
		if (srcRoads.contains(oneOfSrcRoads))
			srcRoads.remove(oneOfSrcRoads);

		resultSet.add(oneOfSrcRoads);
		headNeighborRoads = getNeighborInSrcRoads(oneOfSrcRoads.getState()
				.getConnectedRoads(0, world), srcRoads);
		tailNeighborRoads = getNeighborInSrcRoads(oneOfSrcRoads.getState()
				.getConnectedRoads(1, world), srcRoads);
		if ((headNeighborRoads == null || headNeighborRoads.isEmpty())
				&& (tailNeighborRoads == null || tailNeighborRoads.isEmpty())) {

			return generatePath(id, resultSet, world);
		}

		Road currentHeadRoad = oneOfSrcRoads;
		Road currentTailRoad = oneOfSrcRoads;
		boolean headNull = false, tailNull = false;
		int maxCapacity = 10;
		while ((!headNull || !tailNull) && resultSet.size() <= maxCapacity) {
			if (!headNull) {
				Road cer = chooseExpandRoad(currentHeadRoad, srcRoads,
						resultSet, world);
				if (cer != null) {
					resultSet.add(cer);
					currentHeadRoad = cer;
					srcRoads.remove(cer);
				} else
					headNull = true;
			}
			if (!tailNull) {
				Road cer = chooseExpandRoad(currentTailRoad, srcRoads,
						resultSet, world);
				if (cer != null) {
					resultSet.add(cer);
					currentTailRoad = cer;
					srcRoads.remove(cer);
				} else
					tailNull = true;
			}
		}
		if (!resultSet.isEmpty()) {
			resultPath = generatePath(id, resultSet, world);
		}
		return resultPath;
	}

	static public Path cartographerPath(int id, HashSet<Road> srcRoads,
			StandardWorldModel world) {
		if (srcRoads.isEmpty())
			return null;

		Path ansPath = null;
		ArrayList<Road> preLink = new ArrayList<Road>();
		ArrayList<Road> sufLink = new ArrayList<Road>();
		boolean oneLinkcheck = false;
		ArrayList<Road> mRoads = new ArrayList<Road>(srcRoads);
		int index = 0;
		Road pathSt = mRoads.get(index);

//		System.out.println(pathSt.toString());
//		System.out.println(pathSt.getState().m_roadnode.size());
		while (pathSt.getState().m_roadnode.size() != 2
				&& index < srcRoads.size()-1) {
//			System.out.println(pathSt.toString());
//			System.out.println(pathSt.getState().m_roadnode.size());
			if (pathSt.getState().m_roadnode.size() == 1)
				srcRoads.remove(pathSt);
			pathSt = mRoads.get(++index);
		}
		srcRoads.remove(pathSt);
		if(pathSt.getState().m_roadnode.size()!=2)
			return null;
		// OneLinkSearch
		Road cloneSt = pathSt;
		if(cloneSt.getState().getPointTail(world) == null){
//			System.out.println("111");
			return null;
		}
		
		preLink.add((Road)cloneSt.getState().getPointTail(world).getNeighborhood(world));
		while (!oneLinkcheck) {
			preLink.add(cloneSt);
//			System.out.println(preLink.toString());
			srcRoads.remove(cloneSt);
			cloneSt = (Road) cloneSt.getState().getOtherEndNode(
					preLink.get(preLink.size() - 2), world);
			if(cloneSt == null) return null;
//			System.out.println(cloneSt.toString()+":");
			for(Node nd:cloneSt.getState().m_roadnode)
//				System.out.print(nd.getNeighborhood(world));
//			System.out.println();
			if (cloneSt.getState().m_roadnode.size() != 2)
				oneLinkcheck = true;
		}
		preLink.add(cloneSt);
		srcRoads.remove(cloneSt);
		preLink.remove(pathSt);
		preLink.remove((Road)pathSt.getState().getPointTail(world).getNeighborhood(world));
		
		oneLinkcheck = false;
		cloneSt = pathSt;
		sufLink.add((Road)cloneSt.getState().getPointHead(world).getNeighborhood(world));
		// OtherLinkSearch
		while (!oneLinkcheck) {
			sufLink.add(cloneSt);
			srcRoads.remove(cloneSt);
			cloneSt = (Road) cloneSt.getState().getOtherEndNode(
					sufLink.get(sufLink.size() - 2), world);
//			System.out.println(cloneSt.toString()+":");
			for(Node nd:cloneSt.getState().m_roadnode)
//				System.out.print(nd.getNeighborhood(world));
//			System.out.println();
			if (cloneSt.getState().m_roadnode.size() != 2)
				oneLinkcheck = true;
		}
		sufLink.remove(pathSt);
		sufLink.remove((Road)pathSt.getState().getPointHead(world).getNeighborhood(world));
		sufLink.add(cloneSt);
		srcRoads.remove(cloneSt);
		ansPath = generatePath(id, pathSt, preLink, sufLink, world);
		return ansPath;
	}

	/**
	 * @param id
	 * @param pathSt
	 * @param preLink
	 * @param sufLink
	 * @param world
	 * @return 
	 * By JiangYh 2010-6-16
	 */
	private static Path generatePath(int id, Road pathSt,
			ArrayList<Road> preLink, ArrayList<Road> sufLink,
			StandardWorldModel world) {
		Path ans = new Path(id, world);
		ans.roads.clear();
		ans.roads.addFirst(preLink.get(preLink.size() - 1));
		for (int i = preLink.size() - 2; i >= 0; i--)
			ans.roads.addLast(preLink.get(i));
		ans.roads.addLast(pathSt);
		for (int i = sufLink.size() - 1; i >= 0; i--)
			ans.roads.addLast(sufLink.get(i));
		for (Road rd : ans.roads)
			ans.connectedBuildings.addAll(rd.getState().getConnectedBuildings(
					world));
		for (Road rd : ans.roads)
			ans.nodes.addAll(rd.getState().m_node);
		ans.headNode=ans.roads.getFirst();
		ans.tailNode=ans.roads.getLast();
		return ans;
	}

	static private Road chooseExpandRoad(Road curRoad, HashSet<Road> srcRoads,
			HashSet<Road> haveRoads, StandardWorldModel world) {
		Road result = null;
		int maxDistance = Integer.MIN_VALUE;
		if (srcRoads == null || haveRoads == null
				|| !haveRoads.contains(curRoad)) {
			return null;
		}

		int HEAD = 0, TAIL = 1, i_searchNode = 0;
		boolean haveHeadAncestor = false, haveTailAncestor = false;
		for (Road rd : haveRoads) {
			if (isNeighborRoad(curRoad, rd, HEAD)) {
				haveHeadAncestor = true;
				continue;
			}
			if (isNeighborRoad(curRoad, rd, TAIL))
				haveTailAncestor = true;
		}
		if (!haveHeadAncestor && !haveTailAncestor) {
			for (Road road : srcRoads) {
				if (isNeighborRoad(curRoad, road, HEAD)) {
					i_searchNode = HEAD;
					break;
				} else if (isNeighborRoad(curRoad, road, TAIL)) {
					i_searchNode = TAIL;
					break;
				}
			}
		} else if (!haveHeadAncestor) {
			i_searchNode = HEAD;
		} else if (!haveTailAncestor) {
			i_searchNode = TAIL;
		} else {
			return null;
		}

		HashSet<Road> haveRoadsEcCurrent = new HashSet<Road>(haveRoads);
		haveRoadsEcCurrent.remove(curRoad);
		for (Road rd : srcRoads) {
			if (isNeighborRoad(curRoad, rd, i_searchNode)) {
				boolean isLoop = false;
				for (Road road : haveRoadsEcCurrent) {
					if (isNeighborRoad(rd, road, 0)
							|| isNeighborRoad(rd, road, 1)) {
						isLoop = true;
					}
				}
				if (!isLoop
						&& world.getDistance(curRoad.getID(), rd.getID()) > maxDistance) {
					maxDistance = world
							.getDistance(curRoad.getID(), rd.getID());
					result = rd;
				}
			}
		}
		if (result == null)
			;
		return result;
	}

	static public boolean isNeighborRoad(Road originRoad, Road checkedRoad,
			int HEAD_TAIL) {
		List<Road> neighbors = originRoad.getState().getConnectedRoads(
				HEAD_TAIL, world);
		boolean isNeighbor = false;
		for (Road rd : neighbors) {
			if (rd.getID() == checkedRoad.getID()) {
				isNeighbor = true;
				break;
			}
		}
		return isNeighbor;
	}

	static private HashSet<Road> getNeighborInSrcRoads(List<Road> nei,
			HashSet<Road> srcRoads) {
		HashSet<Road> result = new HashSet<Road>();
		if (nei == null || srcRoads == null) {
			return null;
		}
		HashMap<Integer, Road> M_idToRoad = new HashMap<Integer, Road>();
		for (Road rd : srcRoads) {
			M_idToRoad.put(Integer.valueOf(rd.getID().getValue()), rd);
		}
		for (Road rd : nei) {
			Road returnRD = M_idToRoad.get(Integer.valueOf(rd.getID()
					.getValue()));
			if (returnRD != null && !result.contains(returnRD)) {
				result.add(returnRD);
			}
		}
		return result;
	}

	static public Path generatePath(int id, HashSet<Road> containedRoads,
			StandardWorldModel world) {
		Path path = new Path(id, world);
		path.roads.addAll(containedRoads);
		Road headRoad = null;

		for (Road rd : path.roads) {
			Node head = (Node) rd.getState().getPointHead(world);
			Node tail = (Node) rd.getState().getPointTail(world);
			path.nodes.add(head);
			path.nodes.add(tail);
			boolean haveHeadNeighbor = false, haveTailNeighbor = false;
			for (Road road : path.roads) {
				if (rd != road && isNeighborRoad(rd, road, 0)) {
					haveHeadNeighbor = true;
					continue;
				}
				if (rd != road && isNeighborRoad(rd, road, 1)) {
					haveTailNeighbor = true;
				}
			}
			try {
				if (!haveHeadNeighbor) {
					if (path.headNode == null) {
						path.headNode = (Road) rd.getState()
								.getPointHead(world).getNeighborhood(world);
						headRoad = rd;
					} else if (path.tailNode == null)
						path.tailNode = (Road) rd.getState()
								.getPointHead(world).getNeighborhood(world);
					else {
						throw new Exception();
					}
				}
				if (!haveTailNeighbor) {
					if (path.headNode == null) {
						path.headNode = (Road) rd.getState()
								.getPointTail(world).getNeighborhood(world);
						headRoad = rd;
					} else if (path.tailNode == null)
						path.tailNode = (Road) rd.getState()
								.getPointTail(world).getNeighborhood(world);
					else {
						throw new Exception();
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
			path.connectedBuildings.addAll(rd.getState().getConnectedBuildings(
					world));
		}
		if (headRoad != null) {
			path.roads.clear();
			path.roads.addFirst(headRoad);
			HashSet<Road> conRoads = new HashSet<Road>(containedRoads);
			conRoads.remove(headRoad);

			while (!conRoads.isEmpty()) {
				for (Road rd : conRoads) {
					if (isNeighborRoad(headRoad, rd, 0)
							|| isNeighborRoad(headRoad, rd, 1)) {
						path.roads.addLast(rd);
						break;
					}
				}
				conRoads.remove(path.roads.getLast());
				headRoad = path.roads.getLast();
			}
		} else {
			System.out
					.println("head Road is NULL!   BUT in fact is a circled path!  ");

			path.roads.clear();
			headRoad = containedRoads.iterator().next();
			path.roads.addFirst(headRoad);
			path.headNode = (Road) headRoad.getState().getPointHead(world)
					.getNeighborhood(world);
			path.tailNode = (Road) headRoad.getState().getPointTail(world)
					.getNeighborhood(world);
			HashSet<Road> conRoads = new HashSet<Road>(containedRoads);
			conRoads.remove(headRoad);

			while (!conRoads.isEmpty()) {
				for (Road rd : conRoads) {
					if (isNeighborRoad(headRoad, rd, 0)
							|| isNeighborRoad(headRoad, rd, 1)) {
						path.roads.addLast(rd);
						break;
					}
				}
				conRoads.remove(path.roads.getLast());
				headRoad = path.roads.getLast();
			}
		}

		return path;
	}

	public void setId(int id) {
		this.id = id;
	}

	public Road getHeadNode() {
		return headNode;
	}

	public Road getTailNode() {
		return tailNode;
	}

	public Road getTailRoad() {
		return roads.getFirst();
	}

	public Road getHeadRoad() {
		return roads.getLast();
	}

	public HashSet<Building> getConnectedBuildings() {
		return connectedBuildings;
	}

	public LinkedList<Road> getContainedRoads() {
		return roads;
	}

	public HashSet<Node> getContainedNodes() {
		return nodes;
	}

	public HashSet<StandardEntity> getContainedObjs() {
		HashSet objs = new HashSet<StandardEntity>(roads);
		objs.addAll(nodes);
		return objs;
	}

	public ArrayList<EntityID> toRoute(boolean toTail) {
		if (toTail && headToTailRoute != null)
			return headToTailRoute;
		if (!toTail && tailToHeadRoute != null)
			return tailToHeadRoute;
		Road origin = toTail ? headNode : tailNode;
		ArrayList<EntityID> result = new ArrayList<EntityID>();
		result.add(origin.getID());
		HashSet have = new HashSet();
		have.add(origin);

		boolean hasNext = true;
		while (hasNext) {
			hasNext = false;
			HashSet<StandardEntity> neighbor = new HashSet<StandardEntity>(
					origin.getState().getNeighborhood(world));
			neighbor.removeAll(have);
			for (StandardEntity mlo : neighbor) {
				if (isOnPath(mlo)) {
					result.add(mlo.getID());
					have.add(mlo);
					origin = (Road) mlo;
					hasNext = true;
					break;
				}
			}
		}
		if (toTail)
			headToTailRoute = result;
		else
			tailToHeadRoute = result;
		return result;
	}

	public void reverse() {
		Road head = this.headNode;
		this.headNode = this.tailNode;
		this.tailNode = head;
		ArrayList tempL = null;
		if (tailToHeadRoute != null && headToTailRoute != null) {
			tempL = tailToHeadRoute;
			tailToHeadRoute = headToTailRoute;
			headToTailRoute = tempL;
		} else {
			EntityID tempID = null;
			if (tailToHeadRoute != null)
				for (int i = 0; i < tailToHeadRoute.size() / 2; i++) {
					tempID = tailToHeadRoute.get(i);
					tailToHeadRoute.set(i, tailToHeadRoute.get(tailToHeadRoute
							.size()
							- 1 - i));
					tailToHeadRoute.set(tailToHeadRoute.size() - 1 - i, tempID);
				}
			if (headToTailRoute != null)
				for (int i = 0; i < headToTailRoute.size() / 2; i++) {
					tempID = headToTailRoute.get(i);
					headToTailRoute.set(i, headToTailRoute.get(headToTailRoute
							.size()
							- 1 - i));
					headToTailRoute.set(headToTailRoute.size() - 1 - i, tempID);
				}
		}
		LinkedList<Road> newRoads = new LinkedList<Road>();
		for (int i = this.roads.size() - 1; i >= 0; i--) {
			newRoads.add(this.roads.get(i));
		}
		roads = newRoads;
	}

	// FIXME
	// /**
	// * int[0] is distance; int[1] is id
	// *
	// * @param start
	// * @return
	// */
	// public int[] getMinDistToEnds(Node start) {
	// if (start == null)
	// return null;
	//
	// int[] dist_ID = new int[2];
	//
	// int head = 0, tail = 0;
	//
	// for (Route rt = this.toRoute(false); rt.previous() != null; rt = rt
	// .previous()) {
	// if (rt.end == start)
	// break;
	//
	// if (rt.end instanceof Road) {
	// Road rd = (Road) rt.end;
	// if (rd.getCondition() == Road.CONDITION_BLOCKED) {
	// head += Integer.MAX_VALUE / 2000;
	// } else {
	// head += rd.getLength() / 1000;
	// }
	// }
	// }
	//
	// for (Route rt = this.toRoute(true); rt.previous() != null; rt = rt
	// .previous()) {
	// if (rt.end == start)
	// break;
	//
	// if (rt.end instanceof Road) {
	// Road rd = (Road) rt.end;
	// if (rd.getCondition() == Road.CONDITION_BLOCKED) {
	// tail += Integer.MAX_VALUE / 2000;
	// } else {
	// tail += rd.getLength() / 1000;
	// }
	// }
	// }
	//
	// if (head > tail) {
	// dist_ID[0] = tail;
	// dist_ID[1] = this.tailNode.id;
	// } else {
	// dist_ID[0] = head;
	// dist_ID[1] = this.headNode.id;
	// }
	//
	// return dist_ID;
	// }
	//
	// public HashMap<Integer, Integer> getMinDistToBothEnds(Node start) {
	//
	// boolean fromHead = this.toRoute(false).end.id == this.headNode.id;
	// int head = 0, tail = 0;
	//
	// for (Route rt = this.toRoute(false); rt.previous() != null; rt = rt
	// .previous()) {
	// if (rt.end == start)
	// fromHead = !fromHead;
	//
	// if (fromHead) {
	// if (rt.end instanceof Road) {
	// Road rd = (Road) rt.end;
	// if (rd.getCondition() == Road.CONDITION_BLOCKED) {
	// head += Integer.MAX_VALUE / 2000;
	// } else {
	// head += rd.getLength() / 1000;
	// }
	// }
	// } else {
	// if (rt.end instanceof Road) {
	// Road rd = (Road) rt.end;
	// if (rd.getCondition() == Road.CONDITION_BLOCKED) {
	// tail += Integer.MAX_VALUE / 2000;
	// } else {
	// tail += rd.getLength() / 1000;
	// }
	// }
	// }
	//
	// }
	//
	// HashMap<Integer, Integer> results = new HashMap<Integer, Integer>();
	// results.put(new Integer(headNode.id), new Integer(head));
	// results.put(new Integer(tailNode.id), new Integer(tail));
	//
	// return results;
	// }

	private int m_blockadesTime = -1;
	private int[] m_count;

	/**
	 * int[0] is blocked roads num, and int[1] is unknown roads.
	 * 
	 * @return
	 */
	public int[] getNumOfBlockadesOnPath(int time) {
		if (m_blockadesTime == time)
			return m_count;

		m_blockadesTime = time;

		m_count = new int[2];

		for (Road rd : this.roads) {
			if (rd.getState().getCondition() == RoadState.CONDITION_BLOCKED)
				m_count[0]++;
			if (rd.getState().getCondition() == RoadState.CONDITION_UNKNOWN)
				m_count[1]++;
		}
		return m_count;
	}

	private float m_pathPassValue = -1;

	public float getPathPassValue(StandardWorldModel world) {
		if (m_pathPassValue > 0)
			return m_pathPassValue;
		m_pathPassValue = AgentUtil.distance(world.getState().getCenter().me(),
				this.getCenter())
				/ 2f
				/ (Math.abs(world.getWorldBounds().first().second()
						- world.getWorldBounds().second().second()) + Math
						.abs(world.getWorldBounds().first().first()
								- world.getWorldBounds().second().first()));

		m_pathPassValue *= 10;
		return m_pathPassValue;
	}

	public HashMap<Integer, ArrayList<Road>> reachableRoadsFromEnds;
	public int length = 0;

	public boolean isBlocked = false;
	public boolean isCleared = false;

	public boolean calculateLength(StandardWorldModel world) {
		isBlocked = false;
		reachableRoadsFromEnds = new HashMap<Integer, ArrayList<Road>>();
		if (isCleared == true) {
			length = 0;
			for (Iterator it = roads.iterator(); it.hasNext();) {
				Road r = (Road) it.next();
				length += r.getState().getLength(world);
			}
			ArrayList<Road> tempr = new ArrayList<Road>();
			tempr.addAll(roads);
			reachableRoadsFromEnds.put(tailNode.getID().getValue(),
					new ArrayList<Road>(tempr));
			reachableRoadsFromEnds.put(headNode.getID().getValue(),
					new ArrayList<Road>(tempr));

			return false;
		}

		// 
		ArrayList<Road> heads = new ArrayList<Road>();
		for (EntityID eId : this.toRoute(true)) {
			if (world.getEntity(eId) instanceof Road) {
				Road rd = (Road) world.getEntity(eId);
				if (rd.getState().getCondition() == RoadState.CONDITION_PASSABLE) {
					heads.add(rd);
				} else if (rd.getState().getCondition() == RoadState.CONDITION_BLOCKED)
					break;

			}
		}
		ArrayList<Road> tails = new ArrayList<Road>();
		for (EntityID eId : this.toRoute(false)) {
			if (world.getEntity(eId) instanceof Road) {
				Road rd = (Road) world.getEntity(eId);
				if (rd.getState().getCondition() == RoadState.CONDITION_PASSABLE) {
					tails.add(rd);
				} else if (rd.getState().getCondition() == RoadState.CONDITION_BLOCKED)
					break;

			}
		}

		reachableRoadsFromEnds.put(tailNode.getID().getValue(), tails);
		reachableRoadsFromEnds.put(headNode.getID().getValue(), heads);

		//

		boolean flag = true;
		length = 0;

		for (Iterator it = roads.iterator(); it.hasNext();) {
			Road r = (Road) it.next();
			if (r.getState().getCondition() == RoadState.CONDITION_PASSABLE) {
				length += r.getState().getLength(world);
				continue;
			}
			flag = false;
			if (r.getState().getCondition() == RoadState.CONDITION_BLOCKED) {
				isBlocked = true;
				length += Integer.MAX_VALUE / 2000;
			}

			length += r.getState().getLength(world);
		}

		isCleared = flag;

		return flag;
	}

	// ===================================== for PF evaluation
	// ==========================

}
